"""
Processing of Geometry objects.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import List, Optional

import numpy as np

from . import meshing
from .utils import (
    statistical_outlier_removal,
    eigenvalue_outlier_removal,
    com_cluster_points,
    find_closest_points,
    connected_components,
    envelope_components,
    leiden_clustering,
    dbscan_clustering,
    birch_clustering,
    kmeans_clustering,
)

__all__ = ["GeometryOperations"]


def skeletonize(geometry, method: str = "core", sigma: float = 1.0, **kwargs):
    """
    Extract structural skeleton from point cloud.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    method : {'outer', 'core', 'boundary'}, optional
        Structural feature to extract:
        - 'outer': Outer boundaries
        - 'core': Medial axis/centerline
        - 'boundary': Inner/outer boundaries
        - 'outer_hull': Outer boundaries using a convex hull
    sigma : float, optional
        Gaussian smoothing for Hessian computation.
    **kwargs
        Additional arguments passed to the chosen method.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Decimated geometry.

    Raises
    ------
    ValueError
        If unsupported method is specified.
    """
    from .geometry import Geometry
    from .parametrization import ConvexHull
    from .utils import skeletonize as _skeletonize
    from .utils import points_to_volume, volume_to_points

    method = method.lower()
    methods = ("core", "outer", "boundary", "outer_hull")
    if method not in methods:
        supported = ",".join([f"'{x}'" for x in methods])
        raise ValueError(f"method must be {supported} got '{method}'.")

    skeleton_method = method
    if method == "outer":
        skeleton_method = "boundary"

    points = geometry.points
    if method in ("core", "boundary", "outer"):
        vol, offset = points_to_volume(
            geometry.points, geometry.sampling_rate, use_offset=True
        )
        skeleton = _skeletonize(vol, mode=skeleton_method, sigma=sigma)
        points = volume_to_points(skeleton, geometry.sampling_rate)[0]
        points = np.add(points, offset * geometry.sampling_rate)

    if method in ("outer", "outer_hull"):
        hull = ConvexHull.fit(
            points,
            elastic_weight=0,
            curvature_weight=0,
            volume_weight=0,
            voxel_size=geometry.sampling_rate,
        )
        sample_frac = kwargs.get("sample_fraction", 0.5)
        hull_points = hull.sample(int(sample_frac * points.shape[0]))
        _, indices = find_closest_points(points, hull_points)
        points = points[np.unique(indices)]

    return Geometry(points, sampling_rate=geometry.sampling_rate)


def downsample(geometry, method: str = "radius", **kwargs):
    """
    Reduces point density by removing points based on spatial or random criteria.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    method : str, optional
        Method to use. Options are:
        - 'radius' : Remove points that fall within radius of each other using voxel downsampling
        - 'core' : Replace points that fall within radius of each other by theor centroid.
        - 'number' : Randomly subsample points to target number
        Default is 'radius'.
    **kwargs
        Additional arguments passed to the chosen method:
        - For 'radius': voxel_size parameter for open3d.voxel_down_sample
        - For 'number': size parameter specifying target number of points

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Downsampled geometry.
    """
    from .geometry import Geometry

    method = method.lower()
    points, normals = geometry.points, geometry.normals
    if method.lower() == "radius":
        import open3d as o3d

        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(points)
        pcd.normals = o3d.utility.Vector3dVector(normals)

        pcd = pcd.voxel_down_sample(**kwargs)
        points = np.asarray(pcd.points)
        normals = np.asarray(pcd.normals)
    elif method.lower() == "number":
        size = kwargs.get("size", 1000)
        size = min(size, points.shape[0])
        keep = np.random.choice(range(points.shape[0]), replace=False, size=size)
        points, normals = points[keep], normals[keep]
    elif method.lower() == "center of mass":
        cutoff = kwargs.get("radius", None)
        if cutoff is None:
            cutoff = 4 * np.max(geometry.sampling_rate)
        normals = None
        points = com_cluster_points(points, cutoff)
    else:
        raise ValueError("Supported are 'radius', 'center of mass', and 'number'.")

    return Geometry(points, normals=normals, sampling_rate=geometry._sampling_rate)


def crop(geometry, distance: float, query: np.ndarray, keep_smaller: bool = True):
    """
    Filters points based on their distance to a set of query points.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    distance : float
        Distance threshold for cropping.
    query : np.ndarray
        Points to compute distances to.
    keep_smaller : bool, optional
        If True, keep points closer than distance threshold.
        If False, keep points farther than distance threshold.
        Default is True.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Cropped geometry.
    """
    from mosaic.utils import find_closest_points

    dist, _ = find_closest_points(query, geometry.points, k=1)
    if keep_smaller:
        mask = dist < distance
    else:
        mask = dist >= distance

    return geometry[mask]


def sample(
    geometry,
    sampling: float,
    method: str,
    normal_offset: float = 0.0,
    bidirectional: bool = False,
    **kwargs,
):
    """
    Generates new points by sampling from a fitted parametric model.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    sampling : float
        Sampling rate or number of points to generate.
    method : str
        Sampling method to use. If not "N points", sampling is interpreted
        as a rate and converted to number of points.
    normal_offset : float, optional
        Point offset along normal vector, defaults to 0.0.
    bidirectional : bool, optional
        Draw inward and outward facing points at the same time. This doubles the
        total number of points. Default is False.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Sampled geometry.

    Raises
    ------
    ValueError
        If geometry has no fitted model.
    """
    from .geometry import Geometry

    if (fit := geometry.model) is None:
        return None

    method = method.lower()
    n_samples, extra_kwargs = sampling, {}
    if method != "points":
        n_samples = fit.points_per_sampling(sampling, normal_offset)
        extra_kwargs["mesh_init_factor"] = 5

    # We handle normal offset in sample to ensure equidistant spacing for meshes
    extra_kwargs["normal_offset"] = normal_offset
    points = fit.sample(int(n_samples), **extra_kwargs, **kwargs)
    normals = fit.compute_normal(points)

    if bidirectional:
        extra_kwargs["normal_offset"] = -normal_offset
        new_points = fit.sample(int(n_samples), **extra_kwargs, **kwargs)
        new_normals = -1 * fit.compute_normal(points)
        points = np.concatenate([points, new_points])
        normals = np.concatenate([normals, new_normals])
    return Geometry(points, normals=normals, sampling_rate=geometry.sampling_rate)


def cluster(
    geometry,
    method: str,
    drop_noise: bool = False,
    use_points: bool = True,
    use_normals: bool = False,
    **kwargs,
) -> List:
    """
    Partitions points into clusters using the specified clustering algorithm.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    method : str
        Clustering method to use. Options are:
        - 'DBSCAN' : Density-based clustering
        - 'Birch' : Balanced iterative reducing clustering hierarchy
        - 'K-Means' : K-means clustering
        - 'Connected Components' : Connected component analysis
        - 'Envelope' : Envelope-based clustering
        - 'Leiden' : Leiden community detection
    drop_noise : bool, optional
        If True, drop noise points (label -1) from results.
        Default is False.
    use_points : bool, optional
        If True, use point coordinates for clustering.
        Default is True.
    use_normals : bool, optional
        If True, include normal vectors in clustering features.
        Default is False.
    **kwargs
        Additional arguments passed to the chosen clustering method.

    Returns
    -------
    List[:py:class:`mosaic.geometry.Geometry`]
        List of geometries, one per cluster.

    Raises
    ------
    ValueError
        If unsupported clustering method is specified or too many clusters found.
    """
    _mapping = {
        "DBSCAN": dbscan_clustering,
        "Birch": birch_clustering,
        "K-Means": kmeans_clustering,
        "Connected Components": connected_components,
        "Envelope": envelope_components,
        "Leiden": leiden_clustering,
    }
    func = _mapping.get(method)
    if func is None:
        raise ValueError(
            f"Method must be one of {list(_mapping.keys())}, got '{method}'."
        )

    distance = geometry.sampling_rate
    if method in ("Connected Components", "Envelope", "Leiden"):
        distance = kwargs.pop("distance", -1)
        if np.any(np.array(distance) < 0):
            distance = geometry.sampling_rate
        kwargs["distance"] = distance

    points = np.divide(geometry.points, distance)

    # Prepare feature data for clustering
    data = points
    if use_points and use_normals:
        data = np.concatenate((points, geometry.normals), axis=1)
    elif not use_points and use_normals:
        data = geometry.normals

    labels = func(data, **kwargs)
    unique_labels = np.unique(labels)
    if len(unique_labels) > 10000:
        raise ValueError("Found more than 10k clusters. Try coarser clustering.")

    # Create geometry objects for each cluster
    result_geometries = []
    for label in unique_labels:
        if label == -1 and drop_noise:
            continue
        cluster_geometry = geometry[labels == label]
        result_geometries.append(cluster_geometry)
    return result_geometries


def remove_outliers(geometry, method: str = "statistical", **kwargs):
    """
    Filters out points that are statistical outliers based on local neighborhoods.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    method : str, optional
        Outlier detection method. Options are:
        - 'statistical' : Statistical outlier removal based on neighbor distances
        - 'eigenvalue' : Eigenvalue-based outlier removal
        Default is 'statistical'.
    **kwargs
        Additional parameters for outlier removal method.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry` or None
        Filtered point cloud geometry with outliers removed.
        Returns None if no points remain after filtering.
    """
    func = statistical_outlier_removal
    if method == "eigenvalue":
        func = eigenvalue_outlier_removal
    else:
        if method != "statistical":
            raise ValueError(
                f"Unsupported method '{method}'. Use 'statistical' or 'eigenvalue'."
            )

    mask = func(geometry.points, **kwargs)
    if mask.sum() == 0:
        return None

    return geometry[mask]


def compute_normals(
    geometry, method: str = "Compute", k: int = 15, **kwargs
) -> Optional:
    """
    Calculates normals for points or flips existing normals.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data. This geometry object is modified in-place.
    method : str, optional
        Normal computation method. Options are:
        - 'Compute' : Calculate new normals from point neighborhoods
        - 'Flip' : Flip existing normals (multiply by -1)
        Default is 'Compute'.
    k : int, optional
        Number of neighbors to consider for normal computation.
        Only used when method='Compute'. Default is 15.
    **kwargs
        Additional parameters for normal computation.
    """
    from .utils import compute_normals

    if method == "Flip":
        geometry.normals = geometry.normals * -1
    elif method == "Compute":
        geometry.normals = compute_normals(geometry.points, k=k, **kwargs)
    else:
        raise ValueError(f"Unsupported method '{method}'. Use 'Compute' or 'Flip'.")
    return duplicate(geometry)


def duplicate(geometry, **kwargs):
    """
    Duplicate a geometry.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Geometry to duplicate.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Duplicated geometry.
    """
    return geometry[...]


def visibility(geometry, visible: bool = True, **kwargs):
    """
    Change the visibility of a geometry object

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Geometry to duplicate.
    visible: bool, optional
        Whether the Geometry instance should be visible or not.
    """
    geometry.set_visibility(visible)


def remesh(geometry, method, **kwargs):
    from .geometry import Geometry
    from .parametrization import TriangularMesh

    if not isinstance(mesh := geometry.model, TriangularMesh):
        return None

    method = method.lower()
    mesh = meshing.to_open3d(mesh.vertices.copy(), mesh.triangles.copy())
    if method == "edge length":
        mesh = meshing.remesh(mesh=mesh, **kwargs)
    elif method == "vertex clustering":
        mesh = mesh.simplify_vertex_clustering(**kwargs)
    elif method == "subdivide":
        func = mesh.subdivide_midpoint
        if kwargs.get("smooth"):
            func = mesh.subdivide_loop
        kwargs = {k: v for k, v in kwargs.items() if k != "smooth"}
        mesh = func(**kwargs)
    elif method == "decimation":
        method = kwargs.get("decimation_method", "Triangle Count").lower()
        sampling = kwargs.get("sampling")
        if method == "reduction factor":
            sampling = np.asarray(mesh.triangles).shape[0] // sampling

        if kwargs.get("smooth", False):
            mesh = mesh.simplify_quadric_decimation(int(sampling))
        else:
            import pyfqmr

            simplifier = pyfqmr.Simplify()
            simplifier.setMesh(np.asarray(mesh.vertices), np.asarray(mesh.triangles))
            simplifier.simplify_mesh(
                target_count=int(sampling),
                aggressiveness=5.5,
                preserve_border=True,
                verbose=False,
            )

            vertices, faces, normals = simplifier.getMesh()
            mesh = meshing.to_open3d(vertices, faces)
    else:
        raise ValueError(f"Unsupported remeshing method: {method}")

    return Geometry(sampling_rate=geometry.sampling_rate, model=TriangularMesh(mesh))


def smooth(geometry, method, **kwargs):
    from .geometry import Geometry
    from .parametrization import TriangularMesh

    if not isinstance(mesh := geometry.model, TriangularMesh):
        return None

    method = method.lower()
    mesh = meshing.to_open3d(mesh.vertices.copy(), mesh.triangles.copy())
    n_iterations = int(kwargs.get("n_iterations", 10))
    if method == "taubin":
        mesh = mesh.filter_smooth_taubin(n_iterations)
    elif method == "laplacian":
        mesh = mesh.filter_smooth_laplacian(n_iterations)
    elif method == "average":
        mesh = mesh.filter_smooth_simple(n_iterations)
    else:
        raise ValueError(f"Unsupported smoothing method: {method}")

    return Geometry(sampling_rate=geometry.sampling_rate, model=TriangularMesh(mesh))


def fit(geometry, method, **kwargs):
    from .geometry import Geometry
    from .parametrization import PARAMETRIZATION_TYPE

    _mapping = {
        "Alpha Shape": "convexhull",
        "Ball Pivoting": "mesh",
        "Poisson": "poissonmesh",
        "Cluster Ball Pivoting": "clusterballpivoting",
        "Flying Edges": "flyingedges",
        "Marching Cubes": "marchingcubes",
    }
    method = _mapping.get(method, method)

    if method == "mesh":
        radii = kwargs.get("radii", None)
        try:
            kwargs["radii"] = [float(x) for x in radii.split(",")]
        except Exception as e:
            raise ValueError(f"Incorrect radius specification {radii}.") from e

    kwargs["voxel_size"] = np.max(geometry.sampling_rate)
    if method == "flyingedges" and kwargs.get("distance", -1) != -1:
        kwargs["voxel_size"] = kwargs.get("distance")

    fit_object = PARAMETRIZATION_TYPE.get(method)
    if fit_object is None:
        raise ValueError(f"{method} is not supported ({PARAMETRIZATION_TYPE.keys()}).")

    points = geometry.points

    n = points.shape[0]
    if n < 50 and method not in ["convexhull", "spline"]:
        raise ValueError(f"Insufficient points for fit ({n}<50).")

    fit = fit_object.fit(points, **kwargs)
    if hasattr(fit, "mesh"):
        new_points = fit.vertices
        normals = fit.compute_vertex_normals()
    else:
        new_points = fit.sample(n_samples=1000)
        normals = fit.compute_normal(new_points)

    return Geometry(
        points=new_points,
        normals=normals,
        sampling_rate=geometry.sampling_rate,
        model=fit,
    )


class GeometryOperations:
    """Registry for geometry operation functions."""

    @classmethod
    def register(cls, operation_name: str, func, decorator=None):
        """Register an operation function."""
        if decorator is not None:
            func = decorator(func)
        setattr(cls, operation_name, staticmethod(func))


for operation_name, operation_func in [
    ("skeletonize", skeletonize),
    ("downsample", downsample),
    ("crop", crop),
    ("sample", sample),
    ("cluster", cluster),
    ("remove_outliers", remove_outliers),
    ("compute_normals", compute_normals),
    ("duplicate", duplicate),
    ("visibility", visibility),
    ("remesh", remesh),
    ("smooth", smooth),
    ("fit", fit),
]:
    GeometryOperations.register(operation_name, operation_func)



"""
Implements ProgressDialog, wrapping a ProgressBar in a dialog window.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtCore import Qt, QTimer
from qtpy.QtWidgets import (
    QVBoxLayout,
    QFrame,
    QProgressBar,
    QLabel,
    QDialog,
    QApplication,
    QMessageBox,
)


class ProgressDialog:
    def __init__(self, iterable, title="Processing", parent=None):
        self.total = len(iterable)
        self.iterator = iter(iterable)
        self.current = 0
        self.dialog = QDialog(parent)
        self.dialog.setWindowTitle(title)
        self.dialog.setWindowFlags(
            Qt.WindowType.Dialog | Qt.WindowType.FramelessWindowHint
        )
        self.dialog.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setup_ui()

        QTimer.singleShot(50, self._show_dialog)

    def _show_dialog(self):
        if not self.dialog.isVisible():
            self.dialog.show()
            self.dialog.repaint()
            QApplication.processEvents()

    def setup_ui(self):
        from ..stylesheets import QProgressBar_style

        layout = QVBoxLayout(self.dialog)
        container = QFrame()
        container.setStyleSheet(
            """
            QFrame {
                background-color: white;
                border-radius: 10px;
                border: 1px solid #e5e7eb;
            }
        """
        )
        container_layout = QVBoxLayout(container)
        container_layout.setSpacing(15)
        self.progress_bar = QProgressBar()
        self.progress_bar.setTextVisible(False)
        self.progress_bar.setStyleSheet(QProgressBar_style)
        self.progress_bar.setMaximum(self.total)
        self.progress_bar.setValue(0)

        self.status_label = QLabel()
        self.status_label.setStyleSheet(
            """
            QLabel {
                color: #374151;
                font-size: 13px;
            }
        """
        )
        self.status_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.status_label.setText(f"Processing 0/{self.total}")

        container_layout.addWidget(self.status_label)
        container_layout.addWidget(self.progress_bar)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.addWidget(container)
        self.dialog.setFixedSize(300, 100)

    def update_progress(self, value: int = None):
        """Update progress bar and status label."""
        if value is None:
            value = self.current

        # Ensure dialog is visible before updating
        if not self.dialog.isVisible():
            self.dialog.show()

        self.progress_bar.setValue(value)
        self.status_label.setText(f"Processing {value}/{self.total}")
        QApplication.processEvents()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = next(self.iterator)
            self.current += 1
            self.update_progress()
            return item
        except StopIteration:
            self.update_progress()
            self.dialog.close()
            raise

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.dialog.close()
        if exc_type is not None:
            QMessageBox.warning(None, "Error", str(exc_value))
            # Omit traceback
            return True

    def close(self):
        """Explicitly close the dialog."""
        self.dialog.close()



from functools import partial

import numpy as np
from qtpy.QtWidgets import QWidget, QVBoxLayout

from .. import meshing
from ..widgets.ribbon import create_button
from ..parallel import submit_task, submit_task_batch


def _extract_medial_mesh(mesh_geometry):
    from ..geometry import Geometry
    from ..parametrization import TriangularMesh, PoissonMesh

    model = mesh_geometry.model
    if not model.mesh.is_watertight():
        raise ValueError(
            "Mesh is not watertight. Occupancy-based medial extraction requires "
            "a closed surface to determine interior voxels."
        )

    voxel_size = max(mesh_geometry.sampling_rate)
    skel_points = meshing.medial_mesh(
        model.vertices, model.triangles, voxel_size=voxel_size
    )

    # Much slower than ball pivoting but more robust fits
    model = PoissonMesh.fit(
        positions=skel_points, voxel_size=voxel_size, deldist=2 * voxel_size, depth=7
    )
    geom = Geometry(model=model, sampling_rate=mesh_geometry.sampling_rate)
    geom.change_representation("surface")
    return geom


def _project(
    mesh_geometry,
    geometries,
    use_normals: bool = False,
    invert_normals: bool = False,
    update_normals: bool = False,
):
    from ..geometry import Geometry

    mesh = mesh_geometry.model
    new_geometries, projections, triangles = [], [], []
    for geometry in geometries:
        normals = geometry.normals if use_normals else None
        if normals is not None:
            normals = normals * (-1 if invert_normals else 1)

        kwargs = {
            "points": geometry.points,
            "normals": normals,
            "return_projection": True,
            "return_indices": False,
            "return_triangles": True,
        }
        _, projection, triangle = mesh.compute_distance(**kwargs)

        normals = geometry.normals
        if update_normals:
            normals = mesh.compute_normal(projection)

        projections.append(projection)
        triangles.append(triangle)
        new_geometries.append(
            Geometry(
                points=projection, normals=normals, sampling_rate=geometry.sampling_rate
            )
        )

    if not len(projections):
        return None

    projections = np.concatenate(projections)
    triangles = np.concatenate(triangles)
    new_mesh = mesh.add_projections(projections, triangles, return_indices=False)
    new_mesh = Geometry(model=new_mesh, sampling_rate=mesh_geometry.sampling_rate)

    return new_mesh, new_geometries


class ModelTab(QWidget):
    def __init__(self, cdata, ribbon, legend, **kwargs):
        super().__init__()
        self.cdata = cdata
        self.ribbon = ribbon
        self.legend = legend

        layout = QVBoxLayout(self)
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.ribbon)

    def show_ribbon(self):
        self.ribbon.clear()

        func = self._fit_parallel
        fitting_actions = [
            create_button(
                "Sphere",
                "ph.circle",
                self,
                partial(func, "sphere"),
                "Fit to selected clusters",
            ),
            create_button(
                "Ellipse",
                "ph.link-simple-horizontal-break",
                self,
                partial(func, "ellipsoid"),
                "Fit to selected clusters",
            ),
            create_button(
                "Cylinder",
                "ph.hexagon",
                self,
                partial(func, "cylinder"),
                "Fit to selected clusters",
            ),
            create_button(
                "RBF",
                "ph.dots-nine",
                self,
                partial(func, "rbf"),
                "Fit to selected clusters",
                RBF_SETTINGS,
            ),
            create_button(
                "Mesh",
                "ph.triangle",
                self,
                func,
                "Fit to selected clusters",
                MESH_SETTINGS,
            ),
            create_button(
                "Curve",
                "ph.line-segments",
                self,
                partial(func, "spline"),
                "Fit to selected clusters",
                SPLINE_SETTINGS,
            ),
        ]
        self.ribbon.add_section("Parametrization", fitting_actions)

        mesh_actions = [
            create_button(
                "Sample",
                "ph.broadcast",
                self,
                self._sample_parallel,
                "Generate points from fitted model",
                SAMPLE_SETTINGS,
            ),
        ]
        self.ribbon.add_section("Sampling", mesh_actions)

        mesh_actions = [
            create_button(
                "Repair",
                "ph.wrench",
                self,
                self._repair_mesh,
                "Fix holes and topology issues",
                REPAIR_SETTINGS,
            ),
            create_button(
                "Remesh",
                "ph.arrows-clockwise",
                self,
                self._remesh_parallel,
                "Adjust resolution and quality",
                REMESH_SETTINGS,
            ),
            create_button(
                "Smooth",
                "ph.drop",
                self,
                self._smooth_parallel,
                "Reduce surface noise",
                SMOOTH_SETTINGS,
            ),
            create_button(
                "Project",
                "ph.arrow-line-down",
                self,
                self._project_on_mesh,
                "Project points onto mesh",
                PROJECTION_SETTINGS,
            ),
            create_button(
                "Medial",
                "ph.intersect",
                self,
                self._extract_medial_volumetric,
                "Extract medial surface skeleton from a shell mesh",
            ),
        ]
        self.ribbon.add_section("Mesh Operations", mesh_actions)

    def _default_callback(self, geom):
        from ..parametrization import TriangularMesh

        if isinstance(geom.model, TriangularMesh):
            geom.change_representation("surface")
        self.cdata.models.add(geom)
        self.cdata.models.render()

    def _get_selected_meshes(self):
        from ..parametrization import TriangularMesh

        ret = []
        for geometry in self.cdata.models.get_selected_geometries():
            fit = geometry.model
            if not isinstance(fit, TriangularMesh):
                continue
            ret.append(geometry)
        return ret

    def _repair_mesh(
        self,
        max_hole_size=-1,
        elastic_weight=0,
        curvature_weight=0,
        volume_weight=0,
        boundary_ring=0,
        flip_normals=False,
        **kwargs,
    ):
        from ..parametrization import TriangularMesh

        for geometry in self._get_selected_meshes():
            fit = geometry.model
            fit.mesh.remove_non_manifold_edges()
            fit.mesh.remove_degenerate_triangles()
            fit.mesh.remove_duplicated_triangles()
            fit.mesh.remove_unreferenced_vertices()
            fit.mesh.remove_duplicated_vertices()
            vs, fs = meshing.triangulate_refine_fair(
                vs=fit.vertices,
                fs=fit.triangles,
                alpha=elastic_weight,
                beta=curvature_weight,
                gamma=volume_weight,
                hole_len_thr=max_hole_size,
                n_ring=boundary_ring,
            )
            if flip_normals:
                fs = fs[:, ::-1]
            geom = geometry[...]
            geom._model = TriangularMesh(meshing.to_open3d(vs, fs))
            geom.change_representation("surface")
            self.cdata.models.add(geom)
        return self.cdata.models.render()

    def _fit_parallel(self, method: str, *args, **kwargs):
        from ..operations import GeometryOperations

        # These methods are parallelize and would mess with the worker pool
        tasks, max_concurrent = [], None
        if method in ("Poisson", "Marching Cubes"):
            max_concurrent = 1

        for geometry in self.cdata.data.get_selected_geometries():
            tasks.append(
                {
                    "name": "Parametrization",
                    "func": GeometryOperations.fit,
                    "callback": self._default_callback,
                    "kwargs": {"geometry": geometry, "method": method} | kwargs,
                }
            )
        submit_task_batch(tasks, max_concurrent=max_concurrent)

    def _smooth_parallel(self, method, **kwargs):
        from ..operations import GeometryOperations

        for geometry in self._get_selected_meshes():
            submit_task(
                "Smooth",
                GeometryOperations.smooth,
                self._default_callback,
                geometry,
                method,
                **kwargs,
            )

    def _sample_parallel(self, sampling, method, normal_offset=0.0, **kwargs):
        from ..operations import GeometryOperations

        def _callback(*args, **kwargs):
            self.cdata.data.add(*args, **kwargs)
            self.cdata.data.render()

        for geometry in self.cdata.models.get_selected_geometries():
            submit_task(
                "Sample Fit",
                GeometryOperations.sample,
                _callback,
                geometry,
                method=method,
                sampling=sampling,
                normal_offset=normal_offset,
                **kwargs,
            )

    def _remesh_parallel(self, method, **kwargs):
        from ..operations import GeometryOperations

        for geometry in self._get_selected_meshes():
            submit_task(
                "Remesh",
                GeometryOperations.remesh,
                self._default_callback,
                geometry,
                method,
                **kwargs,
            )

    def _extract_medial_volumetric(self, **kwargs):
        tasks = []
        for geometry in self._get_selected_meshes():
            tasks.append(
                {
                    "name": "Medial Mesh",
                    "func": _extract_medial_mesh,
                    "callback": self._default_callback,
                    "kwargs": {"mesh_geometry": geometry},
                }
            )
        # Setting concurency due to PoissonMesh
        submit_task_batch(tasks, max_concurrent=1)

    def _project_on_mesh(
        self,
        use_normals: bool = False,
        invert_normals: bool = False,
        update_normals: bool = False,
        **kwargs,
    ):
        selected_meshes = self._get_selected_meshes()
        if len(selected_meshes) != 1:
            raise ValueError("Please select one mesh for projection.")

        mesh = selected_meshes[0]
        if mesh.model is None:
            return None

        def _callback(ret):
            new_mesh, new_geometries = ret

            for new_geometry in new_geometries:
                self.cdata.data.add(new_geometry)

            new_mesh.change_representation("surface")
            self.cdata.models.add(new_mesh)
            self.cdata.data.render()
            self.cdata.models.render()

        submit_task(
            "Project",
            _project,
            _callback,
            mesh,
            self.cdata.data.get_selected_geometries(),
            use_normals,
            invert_normals,
            update_normals,
        )


SAMPLE_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Sampling Method",
            "parameter": "method",
            "type": "select",
            "options": ["Points", "Distance"],
            "default": "Distance",
            "notes": "Number of points or average distance between points.",
        },
        {
            "label": "Sampling",
            "parameter": "sampling",
            "type": "float",
            "min": 1,
            "default": 40,
            "notes": "Numerical value for sampling method.",
        },
        {
            "label": "Offset",
            "parameter": "normal_offset",
            "type": "float",
            "default": 0,
            "min": -1e32,
            "notes": "Points are shifted by n times normal vector for particle picking.",
        },
        {
            "label": "Bidirectional",
            "parameter": "bidirectional",
            "type": "boolean",
            "default": False,
            "notes": "Draw inward and outward facing points at the same time. This "
            "doubles the total number of points compared to running sample without "
            "this option set.",
        },
    ],
}

RBF_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Direction",
            "parameter": "direction",
            "type": "select",
            "options": ["xy", "xz", "yz"],
            "default": "xy",
            "description": "Coordinate plane to fit RBF in.",
        },
    ],
}

SPLINE_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Order",
            "parameter": "order",
            "type": "number",
            "default": 3,
            "min": 1,
            "max": 5,
            "description": "Spline order to fit to control points.",
        },
    ],
}

REPAIR_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Elastic Weight",
            "parameter": "elastic_weight",
            "type": "float",
            "default": 0.0,
            "min": -(2**28),
            "description": "Control mesh smoothness and elasticity.",
            "notes": "0 - strong anchoring, 1 - no anchoring, > 1 repulsion.",
        },
        {
            "label": "Curvature Weight",
            "parameter": "curvature_weight",
            "type": "float",
            "default": 0.0,
            "min": -(2**28),
            "description": "Controls propagation of mesh curvature.",
        },
        {
            "label": "Volume Weight",
            "parameter": "volume_weight",
            "type": "float",
            "default": 0.0,
            "min": -(2**28),
            "description": "Controls internal pressure of mesh.",
        },
        {
            "label": "Boundary Ring",
            "parameter": "boundary_ring",
            "type": "number",
            "default": 0,
            "description": "Also optimize n-ring vertices for ill-defined boundaries.",
        },
        {
            "label": "Flexibility",
            "parameter": "anchoring",
            "type": "float_list",
            "default": "1",
            "min": "0",
            "max": "1",
            "description": "Flexibility of inferred vertices. 1 is maximum. Can be "
            "specified for all axes, e.g., 1, or per-axis, e.g., 1;1;0.5.",
        },
        {
            "label": "Hole Size",
            "parameter": "max_hole_size",
            "type": "float",
            "min": -1.0,
            "default": -1.0,
            "description": "Maximum surface area of holes considered for triangulation.",
        },
        {
            "label": "Flip Normals",
            "parameter": "flip_normals",
            "type": "boolean",
            "default": False,
            "description": "Reverse normal direction of the mesh.",
        },
    ],
}


REMESH_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Method",
            "parameter": "method",
            "type": "select",
            "options": [
                "Decimation",
                "Edge Length",
                "Subdivide",
                "Vertex Clustering",
            ],
            "default": "Decimation",
        },
    ],
    "method_settings": {
        "Edge Length": [
            {
                "label": "Edge Length",
                "parameter": "target_edge_length",
                "type": "float",
                "default": 40.0,
                "min": 1e-6,
                "description": "Average edge length to remesh to.",
            },
            {
                "label": "Iterations",
                "parameter": "n_iter",
                "type": "number",
                "default": 100,
                "min": 1,
                "description": "Number of remeshing operations to repeat on the mesh.",
            },
            {
                "label": "Mesh Angle",
                "parameter": "featuredeg",
                "type": "float",
                "default": 30.0,
                "min": 0.0,
                "description": "Minimum angle between faces to preserve the edge feature.",
            },
        ],
        "Vertex Clustering": [
            {
                "label": "Radius",
                "parameter": "voxel_size",
                "type": "float",
                "default": 40.0,
                "min": 1e-6,
                "description": "Radius within which vertices are clustered.",
            },
        ],
        "Decimation": [
            {
                "label": "Method",
                "parameter": "decimation_method",
                "type": "select",
                "options": ["Triangle Count", "Reduction Factor"],
                "default": "Reduction Factor",
                "description": "Choose how to specify the decimation target.",
            },
            {
                "label": "Sampling",
                "parameter": "sampling",
                "type": "float",
                "default": 10,
                "min": 0,
                "description": "Numerical value for reduction method.",
            },
            {
                "label": "Smooth",
                "parameter": "smooth",
                "type": "boolean",
                "default": True,
                "description": "Use quadratic decimation instead of pyfqmr.",
            },
        ],
        "Subdivide": [
            {
                "label": "Iterations",
                "parameter": "number_of_iterations",
                "type": "number",
                "default": 1,
                "min": 1,
                "description": "Number of iterations.",
                "notes": "A single iteration splits each triangle into four triangles.",
            },
            {
                "label": "Smooth",
                "parameter": "smooth",
                "type": "boolean",
                "default": True,
                "description": "Perform smooth midpoint division.",
            },
        ],
    },
}


SMOOTH_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Method",
            "parameter": "method",
            "type": "select",
            "options": [
                "Taubin",
                "Laplacian",
                "Average",
            ],
            "default": "Taubin",
        },
    ],
    "method_settings": {
        "Taubin": [
            {
                "label": "Iterations",
                "parameter": "number_of_iterations",
                "type": "number",
                "default": 10,
                "min": 1,
                "description": "Number of smoothing iterations.",
                "notes": "Taubin filter prevents mesh shrinkage by applying two Laplacian filters with different parameters.",
            },
        ],
        "Laplacian": [
            {
                "label": "Iterations",
                "parameter": "number_of_iterations",
                "type": "number",
                "default": 10,
                "min": 1,
                "description": "Number of smoothing iterations.",
                "notes": "May lead to mesh shrinkage with high iteration counts.",
            },
        ],
        "Average": [
            {
                "label": "Iterations",
                "parameter": "number_of_iterations",
                "type": "number",
                "default": 5,
                "min": 1,
                "description": "Number of smoothing iterations.",
                "notes": "Simplest filter - vertices are replaced by the average of adjacent vertices.",
            },
        ],
    },
}

MESH_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Method",
            "parameter": "method",
            "type": "select",
            "options": [
                "Alpha Shape",
                "Ball Pivoting",
                "Cluster Ball Pivoting",
                "Poisson",
                "Flying Edges",
                "Marching Cubes",
            ],
            "default": "Alpha Shape",
        },
        *REPAIR_SETTINGS["settings"][:5],
    ],
    "method_settings": {
        "Alpha Shape": [
            {
                "label": "Alpha",
                "parameter": "alpha",
                "type": "float",
                "default": 1.0,
                "description": "Alpha-shape parameter.",
                "notes": "Large values yield coarser features.",
            },
            {
                "label": "Scaling Factor",
                "parameter": "resampling_factor",
                "type": "float",
                "default": 12.0,
                "description": "Resample mesh to scaling factor times sampling rate.",
                "notes": "Decrease for creating smoother pressurized meshes.",
            },
            {
                "label": "Distance",
                "parameter": "distance_cutoff",
                "type": "float",
                "default": 2.0,
                "description": "Vertices further than distance time sampling rate are "
                "labled as inferred for subsequent optimization.",
            },
        ],
        "Ball Pivoting": [
            {
                "label": "Radii",
                "parameter": "radii",
                "type": "text",
                "default": "50",
                "description": "Ball radii used for surface reconstruction.",
                "notes": "Use commas to specify multiple radii, e.g. '50,30.5,10.0'.",
            },
            REPAIR_SETTINGS["settings"][5],  # Hole Size
            {
                "label": "Downsample",
                "parameter": "downsample_input",
                "type": "boolean",
                "default": True,
                "description": "Thin input point cloud to core.",
            },
            {
                "label": "Smoothing Steps",
                "parameter": "n_smoothing",
                "type": "number",
                "default": 5,
                "description": "Pre-smoothing steps before fairing.",
                "notes": "Improves repair but less impactful for topolgoy than weights.",
            },
            {
                "label": "Neighbors",
                "parameter": "k_neighbors",
                "type": "number",
                "min": 1,
                "default": 15,
                "description": "Number of neighbors for normal estimations.",
                "notes": "Consider decreasing this value for small point clouds.",
            },
        ],
        "Cluster Ball Pivoting": [
            {
                "label": "Radius",
                "parameter": "radius",
                "type": "float",
                "default": 0.0,
                "max": 100,
                "min": 0.0,
                "description": "Ball radius compared to point cloud box size.",
                "notes": "Default 0 corresponds to an automatically determined radius.",
            },
            {
                "label": "Mesh Angle",
                "parameter": "creasethr",
                "type": "float",
                "min": 0,
                "default": 90.0,
                "description": "Maximum crease angle before stoping ball pivoting.",
            },
            {
                "label": "Smooth Iter",
                "parameter": "smooth_iter",
                "type": "number",
                "min": 1,
                "default": 1,
                "description": "Number of smoothing iterations for normal estimation.",
            },
            {
                "label": "Distance",
                "parameter": "deldist",
                "type": "float",
                "min": -1.0,
                "default": -1.0,
                "description": "Drop vertices distant from input sample points.",
                "notes": "This is post-normalization by the sampling rate.",
            },
            {
                "label": "Neighbors",
                "parameter": "k_neighbors",
                "type": "number",
                "min": 1,
                "default": 15,
                "description": "Number of neighbors for normal estimations.",
                "notes": "Consider decreasing this value for small point clouds.",
            },
        ],
        "Poisson": [
            {
                "label": "Depth",
                "parameter": "depth",
                "type": "number",
                "min": 1,
                "default": 9,
                "description": "Depth of the Octree for surface reconstruction.",
            },
            {
                "label": "Samples",
                "parameter": "samplespernode",
                "type": "float",
                "min": 0,
                "default": 5.0,
                "description": "Minimum number of points per octree node.",
            },
            {
                "label": "Smooth Iter",
                "parameter": "smooth_iter",
                "type": "number",
                "min": 1,
                "default": 1,
                "description": "Number of smoothing iterations for normal estimation.",
            },
            {
                "label": "Pointweight",
                "parameter": "pointweight",
                "type": "float",
                "min": 0,
                "default": 0.1,
                "description": "Interpolation weight of point samples.",
            },
            {
                "label": "Scale",
                "parameter": "scale",
                "type": "float",
                "min": 0,
                "default": 1.2,
                "description": "Ratio between reconstruction and sample cube.",
            },
            {
                "label": "Distance",
                "parameter": "deldist",
                "type": "float",
                "min": -1.0,
                "default": -1.0,
                "description": "Drop vertices further than distance from input.",
            },
            {
                "label": "Neighbors",
                "parameter": "k_neighbors",
                "type": "number",
                "min": 1,
                "default": 15,
                "description": "Number of neighbors for normal estimations.",
                "notes": "Consider decreasing this value for small point clouds.",
            },
        ],
        "Flying Edges": [
            {
                "label": "Distance",
                "parameter": "distance",
                "type": "float",
                "description": "Distance between points to be considered connected.",
                "default": -1.0,
                "min": -1.0,
                "max": 1e32,
                "notes": "Defaults to the sampling rate of the object.",
            },
        ],
        "Marching Cubes": [
            {
                "label": "Simplifcation Factor",
                "parameter": "simplification_factor",
                "type": "number",
                "default": 100,
                "min": 1,
                "description": "Reduce initial mesh by x times the number of triangles.",
            },
            {
                "label": "Workers",
                "parameter": "num_workers",
                "type": "number",
                "default": 8,
                "min": 1,
                "description": "Number of parallel workers to use.",
            },
            {
                "label": "Close Dataset Edges",
                "parameter": "closed_dataset_edges",
                "type": "boolean",
                "default": True,
                "description": "Close mesh at at dataset edges.",
            },
        ],
    },
}


PROJECTION_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Cast Normals",
            "parameter": "use_normals",
            "type": "boolean",
            "default": True,
            "description": "Include normal vectors in raycasting.",
        },
        {
            "label": "Invert Normals",
            "parameter": "invert_normals",
            "type": "boolean",
            "default": False,
            "description": "Invert direction of normal vectors.",
        },
        {
            "label": "Update Normals",
            "parameter": "update_normals",
            "type": "boolean",
            "default": False,
            "description": "Update normal vectors of projection based on the mesh.",
        },
    ],
}



"""
Pipeline execution engine for generating and running batch operations.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import pickle
from uuid import uuid4
from os import makedirs
from os.path import join, exists


import numpy as np

from ..parallel import report_progress
from ..container import DataContainer
from ..formats.writer import write_geometries
from ._utils import strip_filepath, topological_sort, flatten
from ..widgets.container_list import TreeStateData, TreeState


__all__ = ["generate_runs", "execute_run"]


def generate_runs(pipeline_config):
    """
    Generate individual run configurations from a pipeline graph.
    For each input file, creates a linear sequence of operations to execute by
    performing topological sort on the dependency graph.

    Parameters
    ----------
    pipeline_config : dict
        Pipeline configuration containing nodes and metadata

    Returns
    -------
    list
        List of run configurations, where each run is a dict containing:
        - run_id: Unique identifier for this run
        - input_file: Path to input file
        - input_params: Import parameters for this file
        - operations: Ordered list of operations to execute

    Raises
    ------
    ValueError
        If pipeline has cycles or missing node references
    """
    nodes = pipeline_config.get("nodes", [])
    if not nodes:
        return []

    node_map = {node["id"]: node for node in nodes}
    root_nodes = [node for node in nodes if not node.get("inputs", [])]
    if not root_nodes:
        raise ValueError("Pipeline has no root nodes - possible cycle detected")

    import_nodes = [
        node for node in root_nodes if node.get("operation_id") == "import_batch"
    ]

    if not import_nodes:
        raise ValueError("Pipeline must start with an Import Files operation")

    if len(import_nodes) > 1:
        raise ValueError("Pipeline currently supports only one Import Files operation")

    import_node = import_nodes[0]
    input_files = import_node["settings"].get("input_files", [])
    file_parameters = import_node["settings"].get("file_parameters", {})

    if not input_files:
        raise ValueError("No input files specified in Import operation")

    runs = []
    operation_sequence = topological_sort(nodes, node_map, import_node["id"])
    for file_idx, input_file in enumerate(input_files):
        run_id = strip_filepath(input_file)

        operations = []
        for node_id in operation_sequence:
            node = node_map[node_id]

            operation = {
                "operation_id": node["operation_id"],
                "name": node["name"],
                "settings": node["settings"].copy(),
                "group_name": node["settings"].get("group_name", f"{node['name']}_out"),
                "inputs": node.get("inputs", []),
                "save_output": node.get("save_output", True),
                "visible_output": node.get("visible_output", True),
                "node_id": node["id"],
            }

            if node["operation_id"] == "import_batch":
                operation["settings"]["input_file"] = input_file
                operation["settings"]["file_parameters"] = file_parameters.get(
                    input_file, {}
                )

            operations.append(operation)

        run_config = {
            "run_id": run_id,
            "input_file": input_file,
            "input_params": file_parameters.get(input_file, {}),
            "operations": operations,
            "metadata": {
                "file_index": file_idx,
                "total_files": len(input_files),
                "pipeline_version": pipeline_config.get("version", "2.0"),
            },
        }
        runs.append(run_config)

    return runs


def _get_op_spec(operation_id):
    from .operations import OPERATION_CATEGORIES

    for category_data in OPERATION_CATEGORIES.values():
        for op_data in category_data["operations"].values():
            if op_data["id"] == operation_id:
                return op_data.get("input_type"), op_data.get("output_type")
    return None, None


def _load_session(filepath: str):
    from ..formats import open_session

    session = open_session(filepath)
    keys = ("_data_tree", "_models_tree")
    for key in keys:
        tree = session.get(key)
        if tree is None:
            session[key] = TreeStateData()
        elif isinstance(tree, TreeState):
            session[key] = tree.to_tree_state_data()

    return session


def _create_session(filepath: str, parameters: dict):
    from ..formats import open_file

    offset = parameters.get("offset", 0)
    scale = parameters.get("scale", 1)
    sampling = parameters.get("sampling_rate", 1)

    shape = None

    model_container = DataContainer(highlight_color=(0.2, 0.4, 0.8))
    cluster_container = DataContainer()
    for data in open_file(filepath):
        scale_new = np.divide(scale, data.sampling)

        data.vertices = np.subtract(data.vertices, offset, out=data.vertices)
        data.vertices = np.multiply(data.vertices, scale_new, out=data.vertices)
        if data.faces is None:
            cluster_container.add(
                points=data.vertices, normals=data.normals, sampling_rate=sampling
            )
        else:
            from ..meshing import to_open3d
            from ..parametrization import TriangularMesh

            model_container.add(
                model=TriangularMesh(to_open3d(data.vertices, data.faces)),
                sampling_rate=sampling,
            )

        data_shape = np.divide(data.shape, data.sampling)

        if shape is None:
            shape = data_shape
        shape = np.maximum(shape, data_shape)

    metadata = {"shape": shape, "sampling_rate": sampling}
    cluster_container.metadata = metadata.copy()

    data_tree = TreeStateData()
    data_tree.root_items = [x.uuid for x in cluster_container.data]

    model_tree = TreeStateData()
    model_tree.root_items = [x.uuid for x in model_container.data]

    return {
        "shape": shape,
        "_data": cluster_container,
        "_models": model_container,
        "_data_tree": data_tree,
        "_models_tree": model_tree,
    }


def execute_run(run_config: dict, skip_complete: bool = False) -> None:
    """
    Execute a single run configuration.

    Parameters
    ----------
    run_config : dict
        Run configuration generated by :py:meth:`generate_runs`.
    skip_complete : bool, optional
        If True (default), skip execution if all output files already exist.


    Returns
    -------
    str
        Path to the output session file
    """
    from ..operations import GeometryOperations

    current_data = ()
    if skip_complete:
        all_exist, found_export = True, False
        for op in run_config["operations"]:
            if op["operation_id"] not in ("save_session", "export_data"):
                continue
            found_export = True

            settings = op["settings"]
            output_dir = settings.get("output_dir", ".")

            if op["operation_id"] == "save_session":
                output_path = join(output_dir, f"{run_config['run_id']}.pickle")
            elif op["operation_id"] == "export_data":
                output_base = join(output_dir, f"{run_config['run_id']}")
                output_path = f"{output_base}.{settings.get('format')}"

            if not exists(output_path):
                all_exist = False
                break

        if all_exist and found_export:
            print(
                f"Skipping run {run_config['run_id']}: all output files already exist"
            )
            return None

    for idx, op in enumerate(run_config["operations"]):
        op_id = op["operation_id"]
        settings = op["settings"]
        group_name = op["group_name"]

        report_progress(message=op_id, current=idx, total=len(run_config["operations"]))

        # This function gets too much special treatmet
        if op_id == "import_batch":
            input_file = run_config["input_file"]
            try:
                session = _load_session(input_file)
            except Exception:
                session = _create_session(input_file, run_config["input_params"])

            relevant_data = "_data"
            try:
                next_op = run_config["operations"][1]
                input_type, _ = _get_op_spec(next_op["operation_id"])

                if input_type == "model":
                    relevant_data = "_models"

                is_export = next_op["operation_id"] == "export_data"
                if input_type == "any" and is_export:
                    if next_op["settings"]["method"] == "Mesh":
                        relevant_data = "_models"

            except Exception:
                pass

            current_data = session[relevant_data].data

            # i.e. clear the session we just created
            if not op.get("save_output", True):
                current_data = [
                    session[relevant_data].data.pop()
                    for _ in range(len(session[relevant_data].data))
                ]

                for dtype in ("_data", "_models"):
                    metadata = session[dtype].metadata.copy()
                    session[dtype].clear()
                    session[dtype].metadata = metadata

                for dtype in ("_data_tree", "_models_tree"):
                    uuids = session[dtype].get_all_uuids()
                    _ = [session[dtype].remove_uuid(x) for x in uuids]

            # Nothing more to do here
            continue

        if len(current_data) == 0:
            break

        if (func := getattr(GeometryOperations, op_id, None)) is not None:
            # Save some memory over the speedup from the list comprehension
            for i in range(len(current_data)):
                current_data[i] = func(current_data[i], **settings)

        elif op_id == "cluster_select":
            lower_threshold = settings.get("lower_threshold", -1)
            upper_threshold = settings.get("upper_threshold", -1)

            drop = set()
            container, tree = session["_data"], session["_data_tree"]
            for x in current_data:
                keep = True
                n_points = x.get_number_of_points()
                if lower_threshold > 0:
                    keep = keep and n_points > lower_threshold
                if upper_threshold > 0:
                    keep = keep and n_points < upper_threshold

                if not keep:
                    drop.add(x.uuid)
                    tree.remove_uuid(x.uuid)
                    container.remove(x)

            # Data is already in session we just filtered it
            # We still need to update available data for next pipeline step though
            current_data = [x for x in current_data if x.uuid not in drop]
            continue
        elif op_id == "save_session":
            output_dir = settings.get("output_dir", ".")
            makedirs(output_dir, exist_ok=True)
            output_path = join(output_dir, f"{run_config['run_id']}.pickle")

            with open(output_path, "wb") as ofile:
                pickle.dump(session, ofile, protocol=pickle.HIGHEST_PROTOCOL)

            # Do not add current_data to session again
            continue

        elif op_id == "export_data":
            output_dir = settings.get("output_dir", ".")
            makedirs(output_dir, exist_ok=True)

            export_parameters = settings.copy()
            output_path = join(output_dir, f"{run_config['run_id']}")

            # Best guess for the correct shape
            container = session["_data"]
            if (shape := container.metadata.get("shape")) is not None:
                sampling = container.metadata.get("sampling_rate", 1)
                shape = np.rint(np.divide(shape, sampling)).astype(int)

                for key, val in zip(("shape_x", "shape_y", "shape_z"), shape):
                    if key not in export_parameters:
                        export_parameters[key] = val

            # file path will be adapted to carry the correct extension
            write_geometries(
                geometries=current_data,
                file_path=output_path,
                export_parameters=export_parameters
                | {"single_file": True, "include_header": True},
            )

            # Do not add current_data to session again
            continue

        # Some methods return lists of geometry objects
        current_data = flatten(current_data)

        if not op.get("save_output", True):
            continue

        # Keep session data in sync
        input_type, output_type = _get_op_spec(op_id)
        container, tree = session["_data"], session["_data_tree"]
        if output_type == "model":
            container, tree = session["_models"], session["_models_tree"]
            _ = [x.change_representation("surface") for x in current_data]

        if not op.get("visible_output", True):
            _ = [x.set_visibility(False) for x in current_data]

        _ = [container.add(x) for x in current_data]

        group_id = str(uuid4())
        tree.root_items.append(group_id)
        tree.group_names[group_id] = group_name
        tree.groups[group_id] = [x.uuid for x in current_data]
